// POOLSTEUERUNG mit FHEM ver. 2.0
// (C) Alexey Ozerov 2015-2016
// http://baublog.ozerov.de/poolsteuerung-mit-fhem/ 

namespace trick17 {};              // Need this to avoid #ifdef problems with IDE

// *** PH UNIT ***
//#define ENABLE_PH                 // Connected DATA->A1
//#define ENABLE_PHCAL              // Connected ph7: GND->BUTTON->D7, ph4: GND->BUTTON->D8
//#define ENABLE_TX433              // Connected DATA->D2, NB: D11 is also used by VirtualWire lib!

// *** ORP UNIT ***
//#define ENABLE_ORP                // Connected DATA->A2
//#define ENABLE_ORPCAL             // Connected orp: GND->BUTTON->D7
//#define ENABLE_TX433              // Connected DATA->D2, NB: D11 is also used by VirtualWire lib!

// *** ETH UNIT ***
#define ENABLE_PRESSURE           // Connected DATA->A3
#define ENABLE_FLOW               // Connected DATA->D3
#define ENABLE_FLOW2              // Connected DATA->D2
#define ENABLE_T1                 // Connected DATA->D5 + PULL-UP 10K (D4 fails?!)
#define ENABLE_LCD20x4            // Connected I2C SDA->A4 SCL->A5
#define ENABLE_RX433              // Connected DATA->D4, NB: D12 is also used by VirtualWire lib!
#define ENABLE_ETHERNET           // Ethernet W5100 shield

// TBD
// #define ENABLE_T2                 // DHT22 connected DATA->D9

//#define ENABLE_LCD16x2            // Connected LCD shield 16x2

#include <avr/wdt.h>
#include "tools.h"

#ifdef ENABLE_T2
#include "DHT.h"
#define DHTPIN 9     
#define DHTTYPE DHT22 //DHT11, DHT21, DHT22
DHT dht(DHTPIN, DHTTYPE);

#define T2_INTERVAL 1000
unsigned long t2_millis=0;

struct temphum
{ double value;
  double humidity;
} t2;

#endif // ENABLE_T2

#ifdef ENABLE_T1
#include <OneWire.h>            // OneWire-Bibliothek einbinden
#include <DallasTemperature.h>  // DS18B20-Bibliothek einbinden
#define T1_Pin 5
OneWire oneWire(T1_Pin);
DallasTemperature sensors(&oneWire);   // DS18B20 initialisieren

#define T1_INTERVAL 1000
unsigned long t1_millis=0;

struct temperature
{ double value;
} t1;
#endif // ENABLE_T1


#ifdef ENABLE_ETHERNET
#include <SPI.h>
#include <Ethernet.h>
#include <avr/wdt.h>
byte mac[] = {0x00, 0xAA, 0x02, 0x05, 0x19, 0x77};
boolean useEth=1;
IPAddress ip(192, 168, 178, 23);
// IPAddress ip0(0,0,0,0);
EthernetServer server(80);
unsigned long eth_millis=0;
#define ETH_TIMEOUT 60*1000l // Reboot after 1 min without web requests
#endif //ENABLE_ETHERNET

#ifdef ENABLE_LCD20x4
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x3F, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);  // Set the LCD I2C address
//LiquidCrystal_I2C lcd(0x3F, 20, 4);  // Set the LCD I2C address
unsigned long lcd_millis=0;
#define LCD_INTERVAL 200
#endif //ENABLE_LCD20x4

#ifdef ENABLE_LCD16x2
#include <Wire.h>
#include <LiquidCrystal.h>
LiquidCrystal lcd(8, 9, 4, 5, 6, 7);           // select the pins used on the LCD panel
unsigned long lcd_millis=0;
#define LCD_INTERVAL 200
#endif //ENABLE_LCD16x2

#ifdef ENABLE_TX433
#include <VirtualWire.h>
#define TX_INTERVAL 3000
#define TX_INTERVAL_ORP 3100
#define TX_INTERVAL_PH 2900
#define TX433_Pin 2
#define RX433_Pin 11
unsigned long tx_millis=0;
#endif // ENABLE_TX433

#ifdef ENABLE_RX433
#include <VirtualWire.h>
#define TX433_Pin 12
#define RX433_Pin 4
unsigned long rx_millis=0,ph_rx_millis=0,orp_rx_millis=0;
String rx_string = "";
#endif // ENABLE_RX433

#ifdef ENABLE_PHCAL
#define ph7cal_Pin 7
#define ph4cal_Pin 8
#endif // ENABLE_PHCAL

#ifdef ENABLE_ORPCAL
#define orpcal_Pin 7
#endif // ENABLE_ORPCAL

#if defined (ENABLE_PH) || defined(ENABLE_RX433)
#include <EEPROM.h>
#define PH_INTERVAL 53
#define PH_Pin A1
const int ph_magic=0x8077;
const int ph_cal_magic=0xFF44;
const int num_ph_readings = 20;
unsigned long ph_millis=0;

struct ph
{ int magic;
  int reading;
  int reading_min;
  int reading_max;
  double reading_avg;
  double value;
} ph;

struct ph_cal
{ int magic;
  double buffer7;
  double buffer4;
  double reading7;
  double reading4;
  double ratio;
} ph_cal;

int ph_readings[num_ph_readings];         // the readings from the analog input
int ph_index = 0;                         // the index of the current reading
int ph_total = 0;                         // the running total

#endif // ENABLE_PH

#if defined (ENABLE_ORP) || defined(ENABLE_RX433)
#include <EEPROM.h>
#define ORP_INTERVAL 57
#define ORP_Pin A2
const int orp_magic=0x8099;
const int orp_cal_magic=0xFF55;
const int num_orp_readings = 20;
unsigned long orp_millis=0;

struct orp
{ int magic;
  int reading;
  int reading_min;
  int reading_max;
  double reading_avg;
  double value;
} orp;

struct orp_cal
{ int magic;
  double buffer;
  double reading;
  double ratio;
} orp_cal;

int orp_readings[num_orp_readings];        // the readings from the analog input
int orp_index = 0;                         // the index of the current reading
int orp_total = 0;                         // the running total

#endif // ENABLE_ORP

#ifdef ENABLE_PRESSURE
#define PRESSURE_INTERVAL 500
unsigned long pressure_millis=0;
#define PRESSURE_Pin A3

struct pressure
{ int reading;
  double value;
} pressure;

#endif // ENABLE_PRESSURE

#ifdef ENABLE_FLOW
#define FLOW_INTERVAL 1000
unsigned long flow_millis=0;
#define FLOW_Pin       3 // UNO: Pin 2=Interrupt 0, Pin 3=Interrupt 1

volatile int flow_pulse_count;

struct flow
{ int reading;
  double value;
} flow;

// The hall-effect flow sensor outputs approximately 4.5 pulses per second per litre/minute of flow.
// 13.51 is to make 100 for the full flow
const float flow_calibration_factor = 4.5/13.51*60;

#endif // ENABLE_FLOW

#ifdef ENABLE_FLOW2
#define FLOW2_INTERVAL 1007
unsigned long flow2_millis=0;
#define FLOW2_Pin      2 // UNO: Pin 2=Interrupt 0, Pin 3=Interrupt 1

volatile int flow2_pulse_count;

struct flow2
{ int reading;
  double value;
} flow2;

// The hall-effect flow sensor outputs approximately 4.5 pulses per second per litre/minute of flow.
// 13.51 is to make 100 for the full flow
const float flow2_calibration_factor = 4.5/13.51*60*3;

#endif // ENABLE_FLOW2


//******************************************************************************
// SETUP
//******************************************************************************

void setup() {

  Serial.begin(115200);

#ifdef ENABLE_PHCAL
  pinMode(ph7cal_Pin,INPUT_PULLUP);
  pinMode(ph4cal_Pin,INPUT_PULLUP);
#endif // ENABLE_PHCAL

#ifdef ENABLE_ORPCAL
  pinMode(orpcal_Pin,INPUT_PULLUP);
#endif // ENABLE_ORPCAL


#ifdef ENABLE_PH

  EEPROM.get(0, ph_cal);
  if (ph_cal.magic != ph_cal_magic)
  { ph_cal.magic = ph_cal_magic;
    ph_cal.buffer7 = 7.01;                       // For PH7 buffer solution's PH value , 7 or 6.86
    ph_cal.buffer4 = 4.01;                       // For PH4 buffer solution's PH value , 4 or 4.01
    ph_cal.reading7 = 645;                       // PH7 Buffer Solution Reading.
    ph_cal.reading4 = 761;                       // PH4 Buffer Solution Reading.
    ph_cal.ratio = (ph_cal.reading4 - ph_cal.reading7) / (ph_cal.buffer7 - ph_cal.buffer4);                 // Calculate Ph Ratio 
    EEPROM.put(0, ph_cal);
  }

/*
// Init ph calibration settings and readings array
  ph_cal.buffer7 = 7.01;                       // For PH7 buffer solution's PH value , 7 or 6.86
  ph_cal.buffer4 = 4.01;                       // For PH4 buffer solution's PH value , 4 or 4.01
  ph_cal.reading7 = 645;                       // PH7 Buffer Solution Reading.
  ph_cal.reading4 = 761;                       // PH4 Buffer Solution Reading.
  ph_cal.ratio = (ph_cal.reading4 - ph_cal.reading7) / (ph_cal.buffer7 - ph_cal.buffer4);                 // Calculate Ph Ratio 
*/

  for (int i = 0; i < num_ph_readings; i++)
    ph_readings[i] = 0;  

  ph_millis=millis();
#endif // ENABLE_PH

#ifdef ENABLE_ORP

  EEPROM.get(0, orp_cal);
  if (orp_cal.magic != orp_cal_magic)
  { orp_cal.magic = orp_cal_magic;
    orp_cal.buffer  = 465;
    orp_cal.reading = 230;
    orp_cal.ratio = orp_cal.reading / orp_cal.buffer;                 // Calculate ORP Ratio 
    EEPROM.put(0, orp_cal);
  }

/*
// Init orp calibration settings and readings array
  orp_cal.buffer  = 450;
  orp_cal.reading = 230;
  orp_cal.ratio = orp_cal.reading / orp_cal.buffer;                 // Calculate ORP Ratio 
*/

  for (int i = 0; i < num_orp_readings; i++)
    orp_readings[i] = 0;  

#ifdef ENABLE_TX433
  delay(TX_INTERVAL/2);
#endif

  orp_millis=millis();
#endif // ENABLE_ORP

#ifdef ENABLE_PRESSURE
  pressure_millis=millis();
#endif // ENABLE_PRESSURE

#ifdef ENABLE_FLOW
  flow_millis=millis();
  flow_pulse_count  = 0;
  pinMode(FLOW_Pin,INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(FLOW_Pin), flow_pulse_counter, RISING);
#endif // ENABLE_FLOW

#ifdef ENABLE_FLOW2
  flow2_millis=millis();
  flow2_pulse_count  = 0;
  pinMode(FLOW2_Pin,INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(FLOW2_Pin), flow2_pulse_counter, FALLING);
#endif // ENABLE_FLOW2

#ifdef ENABLE_LCD20x4 
  lcd.begin(20,4);         // initialize the lcd for 20 chars 4 lines, turn on backlight
#endif //ENABLE_LCD20x4

#ifdef ENABLE_LCD16x2
  lcd.begin(16,2);         // initialize the lcd for 16 chars 2 lines, turn on backlight
#endif //ENABLE_LCD16x2

#if defined(ENABLE_LCD20x4) || defined(ENABLE_LCD16x2)

//  Serial.println(F("Init LCD..."));

  lcd.backlight();
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print(F("Init Poolduino..."));
  lcd_millis=millis();
#endif //ENABLE_LCD20x4 || ENABLE_LCD16x2

#ifdef ENABLE_TX433
  vw_set_ptt_inverted(true); // Required for RF Link module
  vw_setup(2000);            // Bits per sec
  vw_set_tx_pin(TX433_Pin);
  vw_set_rx_pin(RX433_Pin);
  pinMode(13, OUTPUT);       // Internal LED will blink on transmit
  tx_millis=millis();
#endif // ENABLE_TX433

#ifdef ENABLE_RX433
//  Serial.println("Init RX433...");
  vw_set_ptt_inverted(true); // Required for RF Link module
  vw_setup(2000);            // Bits per sec
  vw_set_tx_pin(TX433_Pin);
  vw_set_rx_pin(RX433_Pin);
  vw_rx_start();             // Start the receiver 
  pinMode(13, OUTPUT);       // Internal LED will blink on transmit
#endif // ENABLE_RX433

#ifdef ENABLE_PH
  show_ph_calibration();
#endif

#ifdef ENABLE_ORP
  show_orp_calibration();
#endif

#ifdef ENABLE_ETHERNET

//  Serial.println(F("Init Ethernet..."));

#if defined(ENABLE_LCD20x4) || defined(ENABLE_LCD16x2)
//  lcd.setCursor(0,1);
//  lcd.print(F("Init Ethernet..."));
#endif //ENABLE_LCD20x4 || ENABLE_LCD16x2
  
  Ethernet.begin(mac,ip);
  server.begin();
//  Serial.print(F("Server is at "));
//  Serial.println(Ethernet.localIP());

#if defined(ENABLE_LCD20x4)
//  lcd.setCursor(0,2);
//  lcd.print(Ethernet.localIP());
#endif //ENABLE_LCD20x4

/*
  if (Ethernet.localIP() == ip0) {
    Serial.println(F("Ethernet failed"));
    useEth=0;
  }
*/

  eth_millis=millis();

#endif //ENABLE_ETHERNET

#ifdef ENABLE_T1

//  Serial.println(F("Init Dallas..."));
  
  sensors.begin();  // DS18B20 starten
//  Serial.print(F("Dallas sensors: "));
//  Serial.println(sensors.getDeviceCount());
  t1_millis=millis();

#if defined(ENABLE_LCD20x4)
//  lcd.setCursor(0,3);
//  lcd.print(F("Dallas sensors:"));
//  lcd.print(sensors.getDeviceCount());
#endif //ENABLE_LCD20x4

#endif // ENABLE_T1

#ifdef ENABLE_T2
  dht.begin();
  t2_millis=millis();
#endif // ENABLE_T2

#if defined(ENABLE_LCD20x4) || defined(ENABLE_LCD16x2)
  delay(1500);
  lcd.clear();
#endif //ENABLE_LCD20x4 || ENABLE_LCD16x2

// Start watchdog 4 sek
  wdt_enable(WDTO_4S);
}

//******************************************************************************
// LOOP
//******************************************************************************

void loop() {

#if defined (ENABLE_PHCAL) && defined (ENABLE_PH)
 if (digitalRead(ph7cal_Pin) != HIGH) 
 { 
   Serial.println(F("ph7 pressed"));
   set_ph7_calibration();
 }

 if (digitalRead(ph4cal_Pin) != HIGH) 
 {
   Serial.println(F("ph4 pressed"));
   set_ph4_calibration();
 }

#endif // ENABLE_PHCAL && ENABLE_PH

#ifdef ENABLE_PH
  if (millis()-ph_millis>PH_INTERVAL)
  { ph_millis=millis();
    get_ph_readings();
  }
#endif // ENABLE_PH

#if defined (ENABLE_ORPCAL) && defined (ENABLE_ORP)
 if (digitalRead(orpcal_Pin) != HIGH)
 {
   Serial.println(F("orp pressed"));
   set_orp_calibration();
 }

#endif // ENABLE_ORPCAL && ENABLE_ORP

#ifdef ENABLE_ORP
  if (millis()-orp_millis>ORP_INTERVAL)
  { orp_millis=millis();
    get_orp_readings();
  }
#endif // ENABLE_ORP

#ifdef ENABLE_PRESSURE
  if (millis()-pressure_millis>PRESSURE_INTERVAL)
  { pressure_millis=millis();
    get_pressure_readings();
  }
#endif // ENABLE_PRESSURE

#ifdef ENABLE_FLOW
  if (millis()-flow_millis>FLOW_INTERVAL)
  { get_flow_readings(); // flow_millis reset inside
  }
#endif // ENABLE_FLOW

#ifdef ENABLE_FLOW2
  if (millis()-flow2_millis>FLOW2_INTERVAL)
  { get_flow2_readings(); // flow2_millis reset inside
  }
#endif // ENABLE_FLOW2

#ifdef ENABLE_T1
  if (millis()-t1_millis>T1_INTERVAL)
  { t1_millis=millis();
    get_temperature_readings();
  }
#endif // ENABLE_T1

#ifdef ENABLE_T2
  if (millis()-t2_millis>T2_INTERVAL)
  { t2_millis=millis();
    get_temperature2_readings();
  }
#endif // ENABLE_T2


#if defined(ENABLE_LCD20x4) || defined(ENABLE_LCD16x2)
  if (millis()-lcd_millis>LCD_INTERVAL)
  { lcd_millis=millis();
    show_lcd_values();
  }
#endif // ENABLE_LCD20x4 || ENABLE_LCD16x2

#if defined(ENABLE_TX433) && defined(ENABLE_ORP)
  if (millis()-tx_millis>TX_INTERVAL_ORP)
  { tx_millis=millis();
    send_tx433();
  }
#elif defined(ENABLE_TX433) && defined(ENABLE_PH)
  if (millis()-tx_millis>TX_INTERVAL_PH)
  { tx_millis=millis();
    send_tx433();
  }
#elif defined(ENABLE_TX433)
  if (millis()-tx_millis>TX_INTERVAL)
  { tx_millis=millis();
    send_tx433();
  }
#endif // ENABLE_TX433 && ENABLE_ORP/PH


#ifdef ENABLE_RX433

  uint8_t buf[VW_MAX_MESSAGE_LEN];
  uint8_t buflen = VW_MAX_MESSAGE_LEN;

  if (vw_get_message(buf, &buflen)) // check to see if anything has been received
  {
/*
    if (buflen>1 && buf[buflen-1]=='\0')
    { Serial.println((char *)buf);
      rx_millis = millis();
      rx_string = String((char *)buf);
    }
    else if (buflen==1 && buf[0]=='\0')
    { Serial.println("RX empty message");
    }
    else 
*/    
    if (buflen==sizeof(ph) && *((int *)buf)==ph_magic)
    { 
      memcpy(&ph,buf,buflen);
      ph_rx_millis = millis();
      
      Serial.print("RX PH:\t");
      Serial.print("ph.reading_avg = " );
      Serial.print(ph.reading_avg);
      Serial.print("\tdelta = " );
      Serial.print(ph.reading_max-ph.reading_min);
      Serial.print("\tph.value = " );
      Serial.print(ph.value);
      Serial.println("");
      
    }
    else if (buflen==sizeof(orp) && *((int *)buf)==orp_magic)
    { 
      memcpy(&orp,buf,buflen);
      orp_rx_millis = millis();

      Serial.print("RX ORP:\t\t");
      Serial.print("orp.reading_avg = " );
      Serial.print(orp.reading_avg);
      Serial.print("\tdelta = " );
      Serial.print(orp.reading_max-orp.reading_min);
      Serial.print("\torp.value = " );
      Serial.print(orp.value);
      Serial.println("");
    }
    else
    {
      Serial.print(F("RX undecodable of "));
      Serial.print(buflen);
      Serial.println(F(" bytes"));
    }
  }

#endif // ENABLE_RX433

#ifdef ENABLE_ETHERNET
  // listen for incoming clients
  EthernetClient client = server.available();
  if (client) {

#ifdef ENABLE_LCD20x4
    lcd.setCursor(19,3);
    lcd.print(F("*"));
#endif
    
    Serial.println(F("Request"));
    // an http request ends with a blank line
    boolean currentLineIsBlank = true;
    while (client.connected()) {
      if (client.available()) {
        char c = client.read();
        Serial.write(c);
        // if you've gotten to the end of the line (received a newline
        // character) and the line is blank, the http request has ended,
        // so you can send a reply
        if (c == '\n' && currentLineIsBlank) {

          eth_millis=millis();   // Save time of the last web access
          
          // send a standard http response header
          client.println(F("HTTP/1.1 200 OK"));
          client.println(F("Content-Type: text/plain"));
          client.println(F("Connection: close"));  // the connection will be closed after completion of the response
          client.println();

#ifdef ENABLE_RX433
          if (millis()-ph_rx_millis<10000 && ph.reading!=0 && ph.reading_max-ph.reading_min<5) {
            client.print(F("ph: "));
            client.print(ph.value,2);
            client.print(" ");
          }
          if (millis()-orp_rx_millis<10000 && orp.reading!=0 && orp.reading_max-orp.reading_min<5) {
            client.print(F("orp: "));
            client.print(orp.value,1);
            client.print(" ");
          }
#endif // ENABLE_RX433

#ifdef ENABLE_PRESSURE
          client.print(F("pressure: "));
          client.print(pressure.value,0);
          client.print(" ");
#endif // ENABLE_PRESSURE

#ifdef ENABLE_FLOW
          client.print(F("flow: "));
          client.print(flow.value,0);
          client.print(" ");
#endif // ENABLE_FLOW

#ifdef ENABLE_T1
          if (t1.value>-127) {
            client.print(F("temperature: "));
            client.print(t1.value,2);
            client.print(" ");
          }
#endif // ENABLE_T1

#ifdef ENABLE_FLOW2
          client.print(F("flow2: "));
          client.print(flow2.value,0);
          client.print(" ");
#endif // ENABLE_FLOW2
          
          client.println("");

          break;
        }
        if (c == '\n') {
          // you're starting a new line
          currentLineIsBlank = true;
        } else if (c != '\r') {
          // you've gotten a character on the current line
          currentLineIsBlank = false;
        }
      }
    }
    // give the web browser time to receive the data
    delay(50);
    // close the connection:
    client.stop();
    Serial.println(F("Request ready"));

#ifdef ENABLE_LCD20x4
    lcd.setCursor(19,3);
    lcd.print(" ");
#endif
  
  }

  if (useEth) {
    Ethernet.maintain();  // Making maintain() leads to reboot if ethernet was missing at startup :-(
    if (millis() - eth_millis > ETH_TIMEOUT) {

#ifdef ENABLE_LCD20x4
//          lcd.setCursor(0,3);
//          lcd.print(F("Going to reset..."));
#endif

      Serial.println(F("Going to reset..."));
      wdt_enable(WDTO_1S); 
      for(;;) { } 

    }
  }
  
#endif //ENABLE_ETHERNET

  wdt_reset();
}

//******************************************************************************
// FUNCTIONS
//******************************************************************************

#ifdef ENABLE_PHCAL
void set_ph7_calibration()
{
  digitalWrite(13, HIGH);
  ph_cal.buffer7 = 7.01;                       // For PH7 buffer solution's PH value , 7 or 6.86
  ph_cal.buffer4 = 4.01;                       // For PH4 buffer solution's PH value , 4 or 4.01
  ph_cal.reading7=ph.reading_avg;
  ph_cal.ratio = (ph_cal.reading4 - ph_cal.reading7) / (ph_cal.buffer7 - ph_cal.buffer4);                 // Calculate new ph Ratio 
  EEPROM.put(0, ph_cal);
  Serial.println(F("New ph7 calibration saved"));
  show_ph_calibration();
  delay(200);
  digitalWrite(13, LOW);
}

void set_ph4_calibration()
{
  digitalWrite(13, HIGH);
  ph_cal.buffer7 = 7.01;                       // For PH7 buffer solution's PH value , 7 or 6.86
  ph_cal.buffer4 = 4.01;                       // For PH4 buffer solution's PH value , 4 or 4.01
  ph_cal.reading4=ph.reading_avg;
  ph_cal.ratio = (ph_cal.reading4 - ph_cal.reading7) / (ph_cal.buffer7 - ph_cal.buffer4);                 // Calculate new ph Ratio 
  EEPROM.put(0, ph_cal);
  Serial.println(F("New ph4 calibration saved"));
  show_ph_calibration();
  delay(200);
  digitalWrite(13, LOW);
}
#endif // ENABLE_PHCAL

#ifdef ENABLE_ORPCAL
void set_orp_calibration()
{
  digitalWrite(13, HIGH);
  orp_cal.buffer  = 465;
  orp_cal.reading=orp.reading_avg;
  orp_cal.ratio = orp_cal.reading / orp_cal.buffer;                 // Calculate new ORP Ratio 
  EEPROM.put(0, orp_cal);
  Serial.println(F("New orp calibration saved"));
  show_orp_calibration();
  delay(200);
  digitalWrite(13, LOW);
}
#endif // ENABLE_ORPCAL

#ifdef ENABLE_PH
void show_ph_calibration()
{
#if defined(ENABLE_LCD20x4) || defined(ENABLE_LCD16x2)
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("PH7:  ");
  lcd.print(ph_cal.reading7);
  lcd.setCursor(0,1);
  lcd.print("PH4:  ");
  lcd.print(ph_cal.reading4);
  lcd.setCursor(0,2);
  lcd.print("Ratio: ");
  lcd.print(ph_cal.ratio);
  delay(3000);
  lcd.clear();
#endif // ENABLE_LCD20x4 || ENABLE_LCD16x2
}

void get_ph_readings() {

// read the analog in value:
  ph.reading = analogRead(PH_Pin);

  ph_total = ph_total - ph_readings[ph_index];
  ph_readings[ph_index] = ph.reading;
  ph_total = ph_total + ph_readings[ph_index];
  ph_index++;
  if (ph_index >= num_ph_readings) ph_index = 0;

  ph.reading_avg = (float)ph_total / num_ph_readings;
  ph.reading_min = minimum(ph_readings,num_ph_readings);
  ph.reading_max = maximum(ph_readings,num_ph_readings);

  ph.value = (ph_cal.reading7 - ph.reading_avg) / ph_cal.ratio + ph_cal.buffer7;    // Calculate PH, no temperature compensation
  ph.magic = ph_magic;

// print the results to the serial monitor:

  Serial.print("ph.reading_avg = " );
  Serial.print(ph.reading_avg);
  Serial.print("\tph.value = " );
  Serial.print(ph.value);
  Serial.println("");

//  Serial.println(ph.reading);
}
#endif // ENABLE_PH


#ifdef ENABLE_ORP
void show_orp_calibration()
{
#if defined(ENABLE_LCD20x4) || defined(ENABLE_LCD16x2)
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("ORP:  ");
  lcd.print(orp_cal.reading);
  lcd.setCursor(0,1);
  lcd.print("Ratio: ");
  lcd.print(orp_cal.ratio);
  delay(3000);
  lcd.clear();
#endif // ENABLE_LCD20x4 || ENABLE_LCD16x2
}

void get_orp_readings() {

// read the analog in value:
  orp.reading = analogRead(ORP_Pin);

  orp_total = orp_total - orp_readings[orp_index];
  orp_readings[orp_index] = orp.reading;
  orp_total = orp_total + orp_readings[orp_index];
  orp_index++;
  if (orp_index >= num_orp_readings) orp_index = 0;

  orp.reading_avg = (float)orp_total / num_orp_readings;
  orp.reading_min = minimum(orp_readings,num_orp_readings);
  orp.reading_max = maximum(orp_readings,num_orp_readings);

  orp.value = orp.reading_avg / orp_cal.ratio;    // Calculate ORP
  orp.magic = orp_magic;

// print the results to the serial monitor:

  Serial.print("orp.reading_avg = " );
  Serial.print(orp.reading_avg);
  Serial.print("\torp.value = " );
  Serial.print(orp.value);
  Serial.println("");
}
#endif // ENABLE_ORP


#ifdef ENABLE_PRESSURE
void get_pressure_readings()
{
  pressure.reading = analogRead(PRESSURE_Pin);
  pressure.value = mapfloat(pressure.reading, 1024*0.5/5.0, 1024*4.5/5.0, 0, 5)*1000;  // tbd: Need some calibration

  Serial.print(F("pressure.reading = " ));
  Serial.print(pressure.reading);
  Serial.print(F("\tpressure.value = " ));
  Serial.print(pressure.value);
  Serial.println(F(""));
}
#endif // ENABLE_PRESSURE

#ifdef ENABLE_FLOW
void get_flow_readings()
{    
  detachInterrupt(digitalPinToInterrupt(FLOW_Pin));
  flow.value = ((1000.0 / (millis() - flow_millis)) * flow_pulse_count) / flow_calibration_factor * 60;

// The hall-effect flow sensor outputs approximately 4.5 pulses per second per litre/minute of flow.
// value l/h = 1000 / 1000 * 750 / 4.5 * 60 = 10000
  
  flow.reading = flow_pulse_count;

  flow_millis=millis();
  flow_pulse_count=0;
  attachInterrupt(digitalPinToInterrupt(FLOW_Pin), flow_pulse_counter, RISING);

  Serial.print(F("flow.reading = " ));
  Serial.print(flow.reading);
  Serial.print(F("\tflow.value = " ));
  Serial.print(flow.value);
  Serial.println(F(""));
}

// Interrupt Service Routine
void flow_pulse_counter()
{
  // Increment the pulse counter
  flow_pulse_count++;
}
#endif // ENABLE_FLOW

#ifdef ENABLE_FLOW2
void get_flow2_readings()
{    
  detachInterrupt(digitalPinToInterrupt(FLOW2_Pin));
  flow2.value = ((1000.0 / (millis() - flow2_millis)) * flow2_pulse_count) / flow2_calibration_factor * 60;
  flow2.reading = flow2_pulse_count;
  
// The hall-effect flow sensor outputs approximately 4.5 pulses per second per litre/minute of flow.
// value l/h = 1000 / 1000 * 750 / 4.5 * 60 = 10000
  
  flow2_millis=millis();
  flow2_pulse_count=0;
  attachInterrupt(digitalPinToInterrupt(FLOW2_Pin), flow2_pulse_counter, FALLING);

  Serial.print(F("flow2.reading = " ));
  Serial.print(flow2.reading);
  Serial.print(F("\tflow2.value = " ));
  Serial.print(flow2.value);
  Serial.println(F(""));
}

// Interrupt Service Routine
void flow2_pulse_counter()
{
  // Increment the pulse counter
  flow2_pulse_count++;
}
#endif // ENABLE_FLOW2


#ifdef ENABLE_T1
void get_temperature_readings()
{
  sensors.requestTemperatures(); // Temperatursensor(en) auslesen
  t1.value = sensors.getTempCByIndex(0);
}
#endif // ENABLE_T1

#ifdef ENABLE_T2
void get_temperature2_readings()
{
  t2.value = dht.readTemperature();
  t2.humidity = dht.readHumidity();
}
#endif // ENABLE_T2

#if defined(ENABLE_LCD20x4) || defined(ENABLE_LCD16x2)

void show_lcd_values()
{

  String str;
  str.reserve(21);

// 1st line
  
#if defined(ENABLE_PH) || defined(ENABLE_RX433)
  lcd.setCursor(0,0);
  lcd.print("PH: ");
//  lcd.print(ph.reading_min);
//  lcd.print("-");
//  lcd.print(ph.reading_max);
//  lcd.print("=");
  lcd.print(ph.reading_avg,1);
  lcd.print("~");
  lcd.print(ph.reading_max-ph.reading_min);
  lcd.print("=");

//  lcd.setCursor(0,1);
//  lcd.print("PH:");
  lcd.print(ph.value);
  lcd.print(" ");
#endif // ENABLE_PH || ENABLE_RX433

// 2nd line

#if defined(ENABLE_ORP) || defined(ENABLE_RX433)
  lcd.setCursor(0,1);
  lcd.print("ORP:");
//  lcd.print(orp.reading_min);
//  lcd.print("-");
//  lcd.print(orp.reading_max);
//  lcd.print("=");
  lcd.print(orp.reading_avg,1);
  lcd.print("~");
  lcd.print(orp.reading_max-orp.reading_min);
  lcd.print("=");

//  lcd.setCursor(0,3);
//  lcd.print("ORP:");
  lcd.print(orp.value);
  lcd.print(" ");
#endif // ENABLE_ORP || ENABLE_RX433

// 3rd line

  str="";

#ifdef ENABLE_T1
  if (t1.value>-127)
  { if (str!="") str+=" ";
    str+="T1:"+String(t1.value,1);
  }
#endif // ENABLE_T1

#ifdef ENABLE_PRESSURE
  if (str!="") str+=" ";
  str+="P:"+String(pressure.value,0);
#endif // ENABLE_PRESSURE

#if defined(ENABLE_T1) || defined(ENABLE_PRESSURE)
  lcd.setCursor(0,2);
  while (str.length()<20) str+=" ";
  lcd.print(str);
#endif // ENABLE_T1 || ENABLE_PRESSURE

// 4th line

  str="";

#ifdef ENABLE_FLOW
  if (str!="") str+=" ";
  str+="F1:"+String(flow.value,0);
#endif // ENABLE_FLOW

#ifdef ENABLE_FLOW2
  if (str!="") str+=" ";
  str+="F2:"+String(flow2.value,0);
#endif // ENABLE_FLOW2

#if defined(ENABLE_FLOW) || defined(ENABLE_FLOW2)
  lcd.setCursor(0,3);
  while (str.length()<19) str+=" ";
  lcd.print(str);
#endif // ENABLE_FLOW || ENABLE_FLOW2

#if defined(ENABLE_ETHERNET)
//  lcd.setCursor(0,3);
//  lcd.print(Ethernet.localIP());
#endif //ENABLE_ETHERNET

}
#endif //ENABLE_LCD20x4 || ENABLE_LCD16x2

#ifdef ENABLE_TX433
void send_tx433()
{ digitalWrite(13, HIGH);   // turn the LED on (HIGH is the voltage level)

#ifdef ENABLE_PH
  vw_send((uint8_t *) &ph, sizeof(ph));
#endif // ENABLE_PH

#ifdef ENABLE_ORP
  vw_send((uint8_t *) &orp, sizeof(orp));
#endif // ENABLE_ORP

/*
  String str="";

#ifdef ENABLE_PH
  if (ph.reading_max-ph.reading_min<=10) // Ignore PH values if noise is too high
  { if (str!="") str+=+", ";
    str+="ph: "+String(ph.value,2);
  }
#endif // ENABLE_PH

#ifdef ENABLE_ORP
  if (orp.reading_max-orp.reading_min<=10) // Ignore ORP values if noise is too high
  { if (str!="") str+=+", ";
    str+="orp: "+String(orp.value,2);
  }
#endif // ENABLE_ORP


#ifdef ENABLE_PRESSURE
  if (str!="") str+=+", ";
  str+="pressure: "+String(pressure.value,2);
#endif // ENABLE_PRESSURE

#ifdef ENABLE_FLOW
  if (str!="") str+=+", ";
  str+="flow: "+String(flow.value,2);
#endif // ENABLE_FLOW

#ifdef ENABLE_T1
  if (t1.value>-127) // Ignore values if sensor detached
  { if (str!="") str+=+", ";
    str+="temperature1: "+String(t1.value,2);
  }
#endif // ENABLE_FLOW

  Serial.print("TX: ");
  Serial.println(str.c_str());

  vw_send((uint8_t *) str.c_str(), str.length()+1);  // c_str() gets internal _buffer of the String object
*/
  
  vw_wait_tx();             // Wait for message to finish
  digitalWrite(13, LOW);    // turn the LED off
  delay(100);               // Wait to relax  
}

#endif // ENABLE_TX433



